const char load_cfg_lua[] =
"-- load_cfg.lua - internal file\n"
"\n"
"local log = require('log')\n"
"local json = require('json')\n"
"local private = require('box.internal')\n"
"local urilib = require('uri')\n"
"local math = require('math')\n"
"local fiber = require('fiber')\n"
"\n"
"-- Function decorator that is used to prevent box.cfg() from\n"
"-- being called concurrently by different fibers.\n"
"local lock = fiber.channel(1)\n"
"local function locked(f)\n"
"    return function(...)\n"
"        lock:put(true)\n"
"        local status, err = pcall(f, ...)\n"
"        lock:get()\n"
"        if not status then\n"
"            error(err)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"--\n"
"-- When feedback is disabled, every single mentioning of it should\n"
"-- be eliminated. Even box.cfg{} should not accept any\n"
"-- 'feedback_*' parameters as valid. This is why they are set to\n"
"-- nil, when the daemon does not exist.\n"
"--\n"
"local function ifdef_feedback(value)\n"
"    return private.feedback_daemon ~= nil and value or nil\n"
"end\n"
"\n"
"local ifdef_feedback_set_params =\n"
"    private.feedback_daemon ~= nil and\n"
"    private.feedback_daemon.set_feedback_params or nil\n"
"\n"
"-- all available options\n"
"local default_cfg = {\n"
"    listen              = nil,\n"
"    memtx_memory        = 256 * 1024 *1024,\n"
"    strip_core          = true,\n"
"    memtx_min_tuple_size = 16,\n"
"    memtx_max_tuple_size = 1024 * 1024,\n"
"    slab_alloc_factor   = 1.05,\n"
"    work_dir            = nil,\n"
"    memtx_dir           = \".\",\n"
"    wal_dir             = \".\",\n"
"\n"
"    vinyl_dir           = '.',\n"
"    vinyl_memory        = 128 * 1024 * 1024,\n"
"    vinyl_cache         = 128 * 1024 * 1024,\n"
"    vinyl_max_tuple_size = 1024 * 1024,\n"
"    vinyl_read_threads  = 1,\n"
"    vinyl_write_threads = 4,\n"
"    vinyl_timeout       = 60,\n"
"    vinyl_run_count_per_level = 2,\n"
"    vinyl_run_size_ratio      = 3.5,\n"
"    vinyl_range_size          = nil, -- set automatically\n"
"    vinyl_page_size           = 8 * 1024,\n"
"    vinyl_bloom_fpr           = 0.05,\n"
"\n"
"    -- logging options are covered by\n"
"    -- a separate log module; they are\n"
"    -- 'log_' prefixed\n"
"\n"
"    io_collect_interval = nil,\n"
"    readahead           = 16320,\n"
"    snap_io_rate_limit  = nil, -- no limit\n"
"    too_long_threshold  = 0.5,\n"
"    wal_mode            = \"write\",\n"
"    wal_max_size        = 256 * 1024 * 1024,\n"
"    wal_dir_rescan_delay= 2,\n"
"    wal_queue_max_size  = 16 * 1024 * 1024,\n"
"    wal_cleanup_delay   = 4 * 3600,\n"
"    force_recovery      = false,\n"
"    replication         = nil,\n"
"    instance_uuid       = nil,\n"
"    replicaset_uuid     = nil,\n"
"    custom_proc_title   = nil,\n"
"    pid_file            = nil,\n"
"    background          = false,\n"
"    username            = nil,\n"
"    coredump            = false,\n"
"    read_only           = false,\n"
"    hot_standby         = false,\n"
"    memtx_use_mvcc_engine = false,\n"
"    checkpoint_interval = 3600,\n"
"    checkpoint_wal_threshold = 1e18,\n"
"    checkpoint_count    = 2,\n"
"    worker_pool_threads = 4,\n"
"    election_mode       = 'off',\n"
"    election_timeout    = 5,\n"
"    replication_timeout = 1,\n"
"    replication_sync_lag = 10,\n"
"    replication_sync_timeout = 300,\n"
"    replication_synchro_quorum = 1,\n"
"    replication_synchro_timeout = 5,\n"
"    replication_connect_timeout = 30,\n"
"    replication_connect_quorum = nil, -- connect all\n"
"    replication_skip_conflict = false,\n"
"    replication_anon      = false,\n"
"    feedback_enabled      = true,\n"
"    feedback_crashinfo    = true,\n"
"    feedback_host         = \"https://feedback.tarantool.io\",\n"
"    feedback_interval     = 3600,\n"
"    net_msg_max           = 768,\n"
"    sql_cache_size        = 5 * 1024 * 1024,\n"
"}\n"
"\n"
"-- cfg variables which are covered by modules\n"
"local module_cfg = {\n"
"    -- logging\n"
"    log                 = log.box_api,\n"
"    log_nonblock        = log.box_api,\n"
"    log_level           = log.box_api,\n"
"    log_format          = log.box_api,\n"
"}\n"
"\n"
"-- types of available options\n"
"-- could be comma separated lua types or 'any' if any type is allowed\n"
"local template_cfg = {\n"
"    listen              = 'string, number',\n"
"    memtx_memory        = 'number',\n"
"    strip_core          = 'boolean',\n"
"    memtx_min_tuple_size  = 'number',\n"
"    memtx_max_tuple_size  = 'number',\n"
"    slab_alloc_factor   = 'number',\n"
"    work_dir            = 'string',\n"
"    memtx_dir            = 'string',\n"
"    wal_dir             = 'string',\n"
"    vinyl_dir           = 'string',\n"
"    vinyl_memory        = 'number',\n"
"    vinyl_cache               = 'number',\n"
"    vinyl_max_tuple_size      = 'number',\n"
"    vinyl_read_threads        = 'number',\n"
"    vinyl_write_threads       = 'number',\n"
"    vinyl_timeout             = 'number',\n"
"    vinyl_run_count_per_level = 'number',\n"
"    vinyl_run_size_ratio      = 'number',\n"
"    vinyl_range_size          = 'number',\n"
"    vinyl_page_size           = 'number',\n"
"    vinyl_bloom_fpr           = 'number',\n"
"\n"
"    log                 = 'module',\n"
"    log_nonblock        = 'module',\n"
"    log_level           = 'module',\n"
"    log_format          = 'module',\n"
"\n"
"    io_collect_interval = 'number',\n"
"    readahead           = 'number',\n"
"    snap_io_rate_limit  = 'number',\n"
"    too_long_threshold  = 'number',\n"
"    wal_mode            = 'string',\n"
"    wal_max_size        = 'number',\n"
"    wal_dir_rescan_delay= 'number',\n"
"    wal_cleanup_delay   = 'number',\n"
"    force_recovery      = 'boolean',\n"
"    replication         = 'string, number, table',\n"
"    instance_uuid       = 'string',\n"
"    replicaset_uuid     = 'string',\n"
"    custom_proc_title   = 'string',\n"
"    pid_file            = 'string',\n"
"    background          = 'boolean',\n"
"    username            = 'string',\n"
"    coredump            = 'boolean',\n"
"    checkpoint_interval = 'number',\n"
"    checkpoint_wal_threshold = 'number',\n"
"    wal_queue_max_size  = 'number',\n"
"    checkpoint_count    = 'number',\n"
"    read_only           = 'boolean',\n"
"    hot_standby         = 'boolean',\n"
"    memtx_use_mvcc_engine = 'boolean',\n"
"    worker_pool_threads = 'number',\n"
"    election_mode       = 'string',\n"
"    election_timeout    = 'number',\n"
"    replication_timeout = 'number',\n"
"    replication_sync_lag = 'number',\n"
"    replication_sync_timeout = 'number',\n"
"    replication_synchro_quorum = 'string, number',\n"
"    replication_synchro_timeout = 'number',\n"
"    replication_connect_timeout = 'number',\n"
"    replication_connect_quorum = 'number',\n"
"    replication_skip_conflict = 'boolean',\n"
"    replication_anon      = 'boolean',\n"
"    feedback_enabled      = ifdef_feedback('boolean'),\n"
"    feedback_crashinfo    = ifdef_feedback('boolean'),\n"
"    feedback_host         = ifdef_feedback('string'),\n"
"    feedback_interval     = ifdef_feedback('number'),\n"
"    net_msg_max           = 'number',\n"
"    sql_cache_size        = 'number',\n"
"}\n"
"\n"
"local function normalize_uri(port)\n"
"    if port == nil or type(port) == 'table' then\n"
"        return port\n"
"    end\n"
"    return tostring(port);\n"
"end\n"
"\n"
"local function normalize_uri_list(port_list)\n"
"    local result = {}\n"
"    if type(port_list) == 'table' then\n"
"        for _, port in ipairs(port_list) do\n"
"            table.insert(result, normalize_uri(port))\n"
"        end\n"
"    else\n"
"        table.insert(result, normalize_uri(port_list))\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"-- options that require special handling\n"
"local modify_cfg = {\n"
"    listen             = normalize_uri,\n"
"    replication        = normalize_uri_list,\n"
"}\n"
"\n"
"local function purge_password_from_uri(uri)\n"
"    local parsed = urilib.parse(uri)\n"
"    if parsed ~= nil and parsed.password ~= nil then\n"
"        return urilib.format(parsed, false)\n"
"    end\n"
"    return uri\n"
"end\n"
"\n"
"local function purge_password_from_uris(uri)\n"
"    if uri == nil then\n"
"        return nil\n"
"    end\n"
"    if type(uri) == 'table' then\n"
"        local new_table = {}\n"
"        for k, v in pairs(uri) do\n"
"            new_table[k] = purge_password_from_uri(v)\n"
"        end\n"
"        return new_table\n"
"    end\n"
"    return purge_password_from_uri(uri)\n"
"end\n"
"\n"
"-- options that require modification for logging\n"
"local log_cfg_option = {\n"
"    replication = purge_password_from_uris,\n"
"}\n"
"\n"
"\n"
"local function check_instance_uuid()\n"
"    if box.cfg.instance_uuid ~= box.info.uuid then\n"
"        box.error(box.error.RELOAD_CFG, 'instance_uuid')\n"
"    end\n"
"end\n"
"\n"
"local function check_replicaset_uuid()\n"
"    if box.cfg.replicaset_uuid ~= box.info.cluster.uuid then\n"
"        box.error(box.error.RELOAD_CFG, 'replicaset_uuid')\n"
"    end\n"
"end\n"
"\n"
"-- dynamically settable options\n"
"--\n"
"-- Note: An option should be in <dynamic_cfg_skip_at_load> table\n"
"-- or should be verified in box_check_config(). Otherwise\n"
"-- load_cfg() may report an error, but box will be configured in\n"
"-- fact.\n"
"local dynamic_cfg = {\n"
"    listen                  = private.cfg_set_listen,\n"
"    replication             = private.cfg_set_replication,\n"
"    log_level               = log.box_api.cfg_set_log_level,\n"
"    log_format              = log.box_api.cfg_set_log_format,\n"
"    io_collect_interval     = private.cfg_set_io_collect_interval,\n"
"    readahead               = private.cfg_set_readahead,\n"
"    too_long_threshold      = private.cfg_set_too_long_threshold,\n"
"    snap_io_rate_limit      = private.cfg_set_snap_io_rate_limit,\n"
"    read_only               = private.cfg_set_read_only,\n"
"    memtx_memory            = private.cfg_set_memtx_memory,\n"
"    memtx_max_tuple_size    = private.cfg_set_memtx_max_tuple_size,\n"
"    vinyl_memory            = private.cfg_set_vinyl_memory,\n"
"    vinyl_max_tuple_size    = private.cfg_set_vinyl_max_tuple_size,\n"
"    vinyl_cache             = private.cfg_set_vinyl_cache,\n"
"    vinyl_timeout           = private.cfg_set_vinyl_timeout,\n"
"    checkpoint_count        = private.cfg_set_checkpoint_count,\n"
"    checkpoint_interval     = private.cfg_set_checkpoint_interval,\n"
"    checkpoint_wal_threshold = private.cfg_set_checkpoint_wal_threshold,\n"
"    wal_queue_max_size      = private.cfg_set_wal_queue_max_size,\n"
"    worker_pool_threads     = private.cfg_set_worker_pool_threads,\n"
"    feedback_enabled        = ifdef_feedback_set_params,\n"
"    feedback_crashinfo      = ifdef_feedback_set_params,\n"
"    feedback_host           = ifdef_feedback_set_params,\n"
"    feedback_interval       = ifdef_feedback_set_params,\n"
"    -- do nothing, affects new replicas, which query this value on start\n"
"    wal_dir_rescan_delay    = function() end,\n"
"    wal_cleanup_delay       = private.cfg_set_wal_cleanup_delay,\n"
"    custom_proc_title       = function()\n"
"        require('title').update(box.cfg.custom_proc_title)\n"
"    end,\n"
"    force_recovery          = function() end,\n"
"    election_mode           = private.cfg_set_election_mode,\n"
"    election_timeout        = private.cfg_set_election_timeout,\n"
"    replication_timeout     = private.cfg_set_replication_timeout,\n"
"    replication_connect_timeout = private.cfg_set_replication_connect_timeout,\n"
"    replication_connect_quorum = private.cfg_set_replication_connect_quorum,\n"
"    replication_sync_lag    = private.cfg_set_replication_sync_lag,\n"
"    replication_sync_timeout = private.cfg_set_replication_sync_timeout,\n"
"    replication_synchro_quorum = private.cfg_set_replication_synchro_quorum,\n"
"    replication_synchro_timeout = private.cfg_set_replication_synchro_timeout,\n"
"    replication_skip_conflict = private.cfg_set_replication_skip_conflict,\n"
"    replication_anon        = private.cfg_set_replication_anon,\n"
"    instance_uuid           = check_instance_uuid,\n"
"    replicaset_uuid         = check_replicaset_uuid,\n"
"    net_msg_max             = private.cfg_set_net_msg_max,\n"
"    sql_cache_size          = private.cfg_set_sql_cache_size,\n"
"}\n"
"\n"
"ifdef_feedback = nil -- luacheck: ignore\n"
"ifdef_feedback_set_params = nil -- luacheck: ignore\n"
"\n"
"--\n"
"-- For some options it is important in which order they are set.\n"
"-- For example, setting 'replication', including self, before\n"
"-- 'listen' makes no sense:\n"
"--\n"
"--     box.cfg{replication = {'localhost:3301'}, listen = 3301}\n"
"--\n"
"-- Replication won't be able to connect to a not being listened\n"
"-- port. In the table below for each option can be set a number.\n"
"-- An option is set before all other options having a bigger\n"
"-- number. Options without a number are installed after others in\n"
"-- an undefined order. The table works for reconfiguration only.\n"
"-- Order of first configuration is hardcoded in C and can't be\n"
"-- changed.\n"
"--\n"
"local dynamic_cfg_order = {\n"
"    listen                  = 100,\n"
"    -- Order of replication_* options does not matter. The only\n"
"    -- rule - apply before replication itself.\n"
"    replication_timeout     = 150,\n"
"    replication_sync_lag    = 150,\n"
"    replication_sync_timeout    = 150,\n"
"    replication_synchro_quorum  = 150,\n"
"    replication_synchro_timeout = 150,\n"
"    replication_connect_timeout = 150,\n"
"    replication_connect_quorum  = 150,\n"
"    replication             = 200,\n"
"    -- Anon is set after `replication` as a temporary workaround\n"
"    -- for the problem, that `replication` and `replication_anon`\n"
"    -- depend on each other. If anon would be configured before\n"
"    -- `replication`, it could lead to a bug, when anon is changed\n"
"    -- from true to false together with `replication`, and it\n"
"    -- would try to deanon the old `replication` before applying\n"
"    -- the new one. This should be fixed when box.cfg is able to\n"
"    -- apply some parameters together and atomically.\n"
"    replication_anon        = 250,\n"
"    -- Cleanup delay should be ignored if replication_anon is set.\n"
"    wal_cleanup_delay       = 260,\n"
"    election_mode           = 300,\n"
"    election_timeout        = 320,\n"
"}\n"
"\n"
"local function sort_cfg_cb(l, r)\n"
"    l = dynamic_cfg_order[l] or math.huge\n"
"    r = dynamic_cfg_order[r] or math.huge\n"
"    return l < r\n"
"end\n"
"\n"
"local dynamic_cfg_skip_at_load = {\n"
"    listen                  = true,\n"
"    memtx_memory            = true,\n"
"    memtx_max_tuple_size    = true,\n"
"    vinyl_memory            = true,\n"
"    vinyl_max_tuple_size    = true,\n"
"    vinyl_cache             = true,\n"
"    vinyl_timeout           = true,\n"
"    too_long_threshold      = true,\n"
"    election_mode           = true,\n"
"    election_timeout        = true,\n"
"    replication             = true,\n"
"    replication_timeout     = true,\n"
"    replication_connect_timeout = true,\n"
"    replication_connect_quorum = true,\n"
"    replication_sync_lag    = true,\n"
"    replication_sync_timeout = true,\n"
"    replication_synchro_quorum = true,\n"
"    replication_synchro_timeout = true,\n"
"    replication_skip_conflict = true,\n"
"    replication_anon        = true,\n"
"    wal_dir_rescan_delay    = true,\n"
"    custom_proc_title       = true,\n"
"    force_recovery          = true,\n"
"    instance_uuid           = true,\n"
"    replicaset_uuid         = true,\n"
"    net_msg_max             = true,\n"
"    readahead               = true,\n"
"}\n"
"\n"
"local function convert_gb(size)\n"
"    return math.floor(size * 1024 * 1024 * 1024)\n"
"end\n"
"\n"
"-- Old to new config translation tables. In case a translation is\n"
"-- not 1-to-1, then a function can be used. It takes 2 parameters:\n"
"-- value of the old option, value of the new if present. It\n"
"-- returns two values - value to replace the old option and to\n"
"-- replace the new one.\n"
"local translate_cfg = {\n"
"    snapshot_count = {'checkpoint_count'},\n"
"    snapshot_period = {'checkpoint_interval'},\n"
"    slab_alloc_arena = {'memtx_memory', function(old)\n"
"        return nil, convert_gb(old)\n"
"    end},\n"
"    slab_alloc_minimal = {'memtx_min_tuple_size'},\n"
"    slab_alloc_maximal = {'memtx_max_tuple_size'},\n"
"    snap_dir = {'memtx_dir'},\n"
"    logger = {'log'},\n"
"    logger_nonblock = {'log_nonblock'},\n"
"    panic_on_snap_error = {'force_recovery', function(old)\n"
"        return nil, not old end\n"
"    },\n"
"    panic_on_wal_error = {'force_recovery', function(old)\n"
"        return nil, not old end\n"
"    },\n"
"    replication_source = {'replication'},\n"
"    rows_per_wal = {'wal_max_size', function(old, new)\n"
"        return old, new\n"
"    end},\n"
"}\n"
"\n"
"-- Upgrade old config\n"
"local function upgrade_cfg(cfg, translate_cfg)\n"
"    if cfg == nil then\n"
"        return {}\n"
"    end\n"
"    local result_cfg = {}\n"
"    for k, v in pairs(cfg) do\n"
"        local translation = translate_cfg[k]\n"
"        if translation ~= nil then\n"
"            local new_key = translation[1]\n"
"            local transform = translation[2]\n"
"            log.warn('Deprecated option %s, please use %s instead', k, new_key)\n"
"            local new_val_orig = cfg[new_key]\n"
"            local old_val, new_val\n"
"            if transform == nil then\n"
"                new_val = v\n"
"            else\n"
"                old_val, new_val = transform(v, new_val_orig)\n"
"            end\n"
"            if new_val_orig ~= nil and\n"
"               new_val_orig ~= new_val then\n"
"                box.error(box.error.CFG, k,\n"
"                          'can not override a value for a deprecated option')\n"
"            end\n"
"            result_cfg[k] = old_val\n"
"            result_cfg[new_key] = new_val\n"
"        else\n"
"            result_cfg[k] = v\n"
"        end\n"
"    end\n"
"    return result_cfg\n"
"end\n"
"\n"
"local function prepare_cfg(cfg, default_cfg, template_cfg,\n"
"                           module_cfg, modify_cfg, prefix)\n"
"    if cfg == nil then\n"
"        return {}\n"
"    end\n"
"    if type(cfg) ~= 'table' then\n"
"        error(\"Error: cfg should be a table\")\n"
"    end\n"
"    -- just pass {.. dont_check = true, ..} to disable check below\n"
"    if cfg.dont_check then\n"
"        return\n"
"    end\n"
"    local readable_prefix = ''\n"
"    if prefix ~= nil and prefix ~= '' then\n"
"        readable_prefix = prefix .. '.'\n"
"    end\n"
"    local new_cfg = {}\n"
"    local module_cfg_backup = {}\n"
"    for k,v in pairs(cfg) do\n"
"        local readable_name = readable_prefix .. k;\n"
"        if template_cfg[k] == nil then\n"
"            box.error(box.error.CFG, readable_name , \"unexpected option\")\n"
"        elseif v == \"\" or v == nil then\n"
"            -- \"\" and NULL = ffi.cast('void *', 0) set option to default value\n"
"            v = default_cfg[k]\n"
"        elseif template_cfg[k] == 'any' then -- luacheck: ignore\n"
"            -- any type is ok\n"
"        elseif type(template_cfg[k]) == 'table' then\n"
"            if type(v) ~= 'table' then\n"
"                box.error(box.error.CFG, readable_name, \"should be a table\")\n"
"            end\n"
"            v = prepare_cfg(v, default_cfg[k], template_cfg[k],\n"
"                            module_cfg[k], modify_cfg[k], readable_name)\n"
"        elseif template_cfg[k] == 'module' then\n"
"            local old_value = module_cfg[k].cfg_get(k, v)\n"
"            module_cfg_backup[k] = old_value or box.NULL\n"
"\n"
"            local ok, msg = module_cfg[k].cfg_set(cfg, k, v)\n"
"            if not ok then\n"
"                -- restore back the old values for modules\n"
"                for module_k, module_v in pairs(module_cfg_backup) do\n"
"                    module_cfg[module_k].cfg_set(nil, module_k, module_v)\n"
"                end\n"
"                box.error(box.error.CFG, readable_name, msg)\n"
"            end\n"
"        elseif (string.find(template_cfg[k], ',') == nil) then\n"
"            -- one type\n"
"            if type(v) ~= template_cfg[k] then\n"
"                box.error(box.error.CFG, readable_name, \"should be of type \"..\n"
"                    template_cfg[k])\n"
"            end\n"
"        else\n"
"            local good_types = string.gsub(template_cfg[k], ' ', '');\n"
"            if (string.find(',' .. good_types .. ',', ',' .. type(v) .. ',') == nil) then\n"
"                box.error(box.error.CFG, readable_name, \"should be one of types \"..\n"
"                    template_cfg[k])\n"
"            end\n"
"        end\n"
"        if modify_cfg ~= nil and type(modify_cfg[k]) == 'function' then\n"
"            v = modify_cfg[k](v)\n"
"        end\n"
"        new_cfg[k] = v\n"
"    end\n"
"    return new_cfg\n"
"end\n"
"\n"
"local function apply_default_cfg(cfg, default_cfg, module_cfg)\n"
"    for k,v in pairs(default_cfg) do\n"
"        if cfg[k] == nil then\n"
"            cfg[k] = v\n"
"        elseif type(v) == 'table' then\n"
"            apply_default_cfg(cfg[k], v)\n"
"        end\n"
"    end\n"
"    for k in pairs(module_cfg) do\n"
"        if cfg[k] == nil then\n"
"            cfg[k] = module_cfg[k].cfg_get(k)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- Return true if two configurations are equivalent.\n"
"local function compare_cfg(cfg1, cfg2)\n"
"    if type(cfg1) ~= type(cfg2) then\n"
"        return false\n"
"    end\n"
"    if type(cfg1) ~= 'table' then\n"
"        return cfg1 == cfg2\n"
"    end\n"
"    if #cfg1 ~= #cfg2 then\n"
"        return false\n"
"    end\n"
"    for k, v in ipairs(cfg1) do\n"
"        if v ~= cfg2[k] then\n"
"            return false\n"
"        end\n"
"    end\n"
"    return true\n"
"end\n"
"\n"
"local function reload_cfg(oldcfg, cfg)\n"
"    cfg = upgrade_cfg(cfg, translate_cfg)\n"
"    local newcfg = prepare_cfg(cfg, default_cfg, template_cfg,\n"
"                               module_cfg, modify_cfg)\n"
"    local ordered_cfg = {}\n"
"    -- iterate over original table because prepare_cfg() may store NILs\n"
"    for key, val in pairs(cfg) do\n"
"        if dynamic_cfg[key] == nil and oldcfg[key] ~= val then\n"
"            box.error(box.error.RELOAD_CFG, key);\n"
"        end\n"
"        table.insert(ordered_cfg, key)\n"
"    end\n"
"    table.sort(ordered_cfg, sort_cfg_cb)\n"
"    for _, key in pairs(ordered_cfg) do\n"
"        local val = newcfg[key]\n"
"        local oldval = oldcfg[key]\n"
"        if not compare_cfg(val, oldval) then\n"
"            rawset(oldcfg, key, val)\n"
"            if not pcall(dynamic_cfg[key]) then\n"
"                rawset(oldcfg, key, oldval) -- revert the old value\n"
"                return box.error() -- re-throw\n"
"            end\n"
"            if log_cfg_option[key] ~= nil then\n"
"                val = log_cfg_option[key](val)\n"
"            end\n"
"            log.info(\"set '%s' configuration option to %s\", key,\n"
"                json.encode(val))\n"
"        end\n"
"    end\n"
"    if type(box.on_reload_configuration) == 'function' then\n"
"        box.on_reload_configuration()\n"
"    end\n"
"end\n"
"\n"
"local box_cfg_guard_whitelist = {\n"
"    error = true;\n"
"    internal = true;\n"
"    index = true;\n"
"    session = true;\n"
"    tuple = true;\n"
"    runtime = true;\n"
"    ctl = true,\n"
"    NULL = true;\n"
"};\n"
"\n"
"-- List of box members that requires full box loading.\n"
"local box_restore_after_full_load_list = {\n"
"    execute = true,\n"
"}\n"
"\n"
"local box = require('box')\n"
"-- Move all box members except the whitelisted to box_configured\n"
"local box_configured = {}\n"
"for k, v in pairs(box) do\n"
"    box_configured[k] = v\n"
"    if not box_cfg_guard_whitelist[k] then\n"
"        box[k] = nil\n"
"    end\n"
"end\n"
"\n"
"setmetatable(box, {\n"
"    __index = function(table, index) -- luacheck: no unused args\n"
"        error(debug.traceback(\"Please call box.cfg{} first\"))\n"
"        error(\"Please call box.cfg{} first\")\n"
"     end\n"
"})\n"
"\n"
"-- Whether box is loaded.\n"
"--\n"
"-- `false` when box is not configured or when the initialization\n"
"-- is in progress.\n"
"--\n"
"-- `true` when box is configured.\n"
"--\n"
"-- Use locked() wrapper to obtain reliable results.\n"
"local box_is_configured = false\n"
"\n"
"local function load_cfg(cfg)\n"
"    -- A user may save box.cfg (this function) before box loading\n"
"    -- and call it afterwards. We should reconfigure box in the\n"
"    -- case.\n"
"    if box_is_configured then\n"
"        reload_cfg(box.cfg, cfg)\n"
"        return\n"
"    end\n"
"\n"
"    cfg = upgrade_cfg(cfg, translate_cfg)\n"
"    cfg = prepare_cfg(cfg, default_cfg, template_cfg,\n"
"                      module_cfg, modify_cfg)\n"
"    apply_default_cfg(cfg, default_cfg, module_cfg);\n"
"    -- Save new box.cfg\n"
"    box.cfg = cfg\n"
"    if not pcall(private.cfg_check)  then\n"
"        box.cfg = locked(load_cfg) -- restore original box.cfg\n"
"        return box.error() -- re-throw exception from check_cfg()\n"
"    end\n"
"\n"
"    -- NB: After this point the function should not raise an\n"
"    -- error.\n"
"    --\n"
"    -- This is important to have right <box_is_configured> (this\n"
"    -- file) and <is_box_configured> (box.cc) values.\n"
"    --\n"
"    -- It also would be counter-intuitive to receive an error from\n"
"    -- box.cfg({<...>}), but find that box is actually configured.\n"
"\n"
"    -- Restore box members after initial configuration.\n"
"    for k, v in pairs(box_configured) do\n"
"        if not box_restore_after_full_load_list[k] then\n"
"            box[k] = v\n"
"        end\n"
"    end\n"
"    setmetatable(box, nil)\n"
"    box.cfg = setmetatable(cfg,\n"
"        {\n"
"            __newindex = function(table, index) -- luacheck: no unused args\n"
"                error('Attempt to modify a read-only table')\n"
"            end,\n"
"            __call = locked(reload_cfg),\n"
"        })\n"
"\n"
"    -- This call either succeeds or calls panic() / exit().\n"
"    private.cfg_load()\n"
"\n"
"    -- This block does not raise an error: all necessary checks\n"
"    -- already performed in private.cfg_check(). See <dynamic_cfg>\n"
"    -- comment.\n"
"    for key, fun in pairs(dynamic_cfg) do\n"
"        local val = cfg[key]\n"
"        if val ~= nil then\n"
"            if not dynamic_cfg_skip_at_load[key] then\n"
"                fun()\n"
"            end\n"
"            if not compare_cfg(val, default_cfg[key]) then\n"
"                if log_cfg_option[key] ~= nil then\n"
"                    val = log_cfg_option[key](val)\n"
"                end\n"
"                log.info(\"set '%s' configuration option to %s\", key, json.encode(val))\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    if not box.cfg.read_only and not box.cfg.replication and\n"
"       not box.error.injection.get('ERRINJ_AUTO_UPGRADE') then\n"
"        box.schema.upgrade{auto = true}\n"
"    end\n"
"\n"
"    -- Restore box members that requires full box loading.\n"
"    for k, v in pairs(box_configured) do\n"
"        if box_restore_after_full_load_list[k] then\n"
"            box[k] = v\n"
"        end\n"
"    end\n"
"    box_configured = nil\n"
"\n"
"    box_is_configured = true\n"
"\n"
"    -- Check if schema version matches Tarantool version and print\n"
"    -- warning if it's not (in case user forgot to call\n"
"    -- box.schema.upgrade()).\n"
"    local needs, schema_version_str = private.schema_needs_upgrade()\n"
"    if needs then\n"
"        local msg = string.format(\n"
"            'Your schema version is %s while Tarantool %s requires a more'..\n"
"            ' recent schema version. Please, consider using box.'..\n"
"            'schema.upgrade().', schema_version_str, box.info.version)\n"
"        log.warn(msg)\n"
"    end\n"
"end\n"
"box.cfg = locked(load_cfg)\n"
"\n"
"--\n"
"-- This makes possible do box.execute without calling box.cfg\n"
"-- manually. The load_cfg() call overwrites box.execute, see\n"
"-- <box_configured> variable.\n"
"--\n"
"-- box.execute is <box_load_and_execute> when box is not loaded,\n"
"-- <lbox_execute> otherwise. <box_load_and_execute> loads box and\n"
"-- calls <lbox_execute>.\n"
"--\n"
"local box_load_and_execute\n"
"box_load_and_execute = function(...)\n"
"    -- Configure box if it is not configured, no-op otherwise (not\n"
"    -- a reconfiguration).\n"
"    --\n"
"    -- We should check whether box is configured, because a user\n"
"    -- may save <box_load_and_execute> function before box loading\n"
"    -- and call it afterwards.\n"
"    if not box_is_configured then\n"
"        locked(function()\n"
"            -- Re-check, because after releasing of the lock the\n"
"            -- box state may be changed. We should call load_cfg()\n"
"            -- only once.\n"
"            if not box_is_configured then\n"
"                load_cfg()\n"
"            end\n"
"        end)()\n"
"    end\n"
"\n"
"    -- At this point box should be configured and box.execute()\n"
"    -- function should be replaced with lbox_execute().\n"
"    assert(type(box.cfg) ~= 'function')\n"
"    assert(box.execute ~= box_load_and_execute)\n"
"\n"
"    return box.execute(...)\n"
"end\n"
"box.execute = box_load_and_execute\n"
"\n"
"-- gh-810:\n"
"-- hack luajit default cpath\n"
"-- commented out because we fixed luajit to build properly, see\n"
"-- https://github.com/luajit/luajit/issues/76\n"
"-- local format = require('tarantool').build.mod_format\n"
"-- package.cpath = package.cpath:gsub(\n"
"--     '\?.so', '\?.' .. format\n"
"-- ):gsub('loadall.so', 'loadall.' .. format)\n"
""
;
